import { finalize } from 'rxjs/operators';
export class Cycler {
    constructor() {
        this._destructors = [];
        this._transientDestructors = {};
    }
    /**
     * Gets an amount of destructors currently in the queue.
     */
    getDestructorsCount() {
        return this._destructors.length;
    }
    /**
     * Disposes the Cycler causing all it's dependant managed
     * entities to be disposed. This method will be called
     * automatically by Angular if the Cycler is properly
     * provided.
     */
    dispose() {
        for (let i = 0; i < this._destructors.length; i++) {
            const destructor = this._destructors[i];
            destructor();
        }
        this._destructors = [];
    }
    /**
     * Angular lifecycle hook implementation. Expected to be
     * called by Angular automatically. There are two ways to
     * achieve that:
     *
     * ### Option 1: provide the Cycler object:
     * ```ts
     * import { Cycler } from 'ng-cycler';
     *
     * @Component({
     *   // ...
     *   providers: [Cycler] // provide component scope-related Cycler instance
     * })
     * export class MyComponent {
     *     constructor(private readonly _cycler: Cycler){
     *         // use _cycler wherever needed
     *     }
     * }
     * ```
     * ### Option 2: extend your component from Cycler:
     *
     * ```ts
     * import { Cycler } from 'ng-cycler';
     *
     * export class MyComponent extends Cycler {
     *     // use this. to call Cycler-related methods
     * }
     * ```
     */
    ngOnDestroy() {
        this.dispose();
    }
    /**
     * Adds a Destructor to the disposal queue to be called once the Cycler
     * is finalized by Angular.
     *
     * @param destructor to be added to the disposal queue
     * @returns disposable that can be used to run the destructor at any
     * time by calling `dispose` method. In case of `dispose` forced
     * manually it will be excluded from the main disposal queue.
     */
    manageDestructor(destructor) {
        this._destructors.push(destructor);
        return {
            dispose: () => {
                const destructorIndex = this._destructors.indexOf(destructor);
                if (destructorIndex >= 0) {
                    destructor();
                    this._destructors.splice(destructorIndex, 1);
                }
            }
        };
    }
    /**
     * Adds a Disposable object to the disposal queue. The logic is the same
     * as for `manageDestructor` but using a Disposable instead of Destructor.
     *
     * See `manageDestructor` for details.
     *
     * @param disposable to be added to the disposal queue
     *
     * @returns disposable that wraps the original disposable. This new Disposable
     * can be used to finalize the current one and remove it from the queue at the
     * same time.
     */
    manageDisposable(disposable) {
        return this.manageDestructor(() => disposable.dispose());
    }
    manageSubscription(subscription) {
        return this.manageDestructor(() => subscription.unsubscribe());
    }
    manage(dependency) {
        const disposable = dependency;
        if (disposable.dispose !== undefined) {
            return this.manageDisposable(disposable);
        }
        const subscription = dependency;
        if (subscription.unsubscribe !== undefined) {
            return this.manageSubscription(subscription);
        }
        return this.manageDestructor(dependency);
    }
    manageTransient(code, destructor) {
        if (this._transientDestructors[code]) {
            this._transientDestructors[code].dispose();
        }
        const result = this.manageDestructor(destructor);
        this._transientDestructors[code] = result;
        return result;
    }
    manageTransientSubscription(code, subscription) {
        return this.manageTransient(code, () => {
            subscription.unsubscribe();
        });
    }
    manageSubscriptions(...subscriptions) {
        for (const subscription of subscriptions) {
            this.manageSubscription(subscription);
        }
    }
    managedSubscribe(observable, next) {
        let destructor = null;
        const subscription = observable
            .pipe(finalize(() => {
            if (destructor !== null) {
                this._destructors = this._destructors.filter(d => d !== destructor);
            }
        }))
            .subscribe(next);
        destructor = () => subscription.unsubscribe();
        return this.manageDestructor(destructor);
    }
}
//# sourceMappingURL=cycler.js.map